local PrivateChat = LibStub("AceAddon-3.0"):NewAddon("PrivateChat", "AceConsole-3.0", "AceHook-3.0", "AceComm-3.0", "AceSerializer-3.0")
local L = LibStub("AceLocale-3.0"):GetLocale("PrivateChat", false)

local AES = LibStub("AES-1.0")

-- Breue:bumped these 11/28
local PROTOCOL_VERSION = 3
local DATABASE_VERSION = 3

local Channel = {}
local mt = { __index = Channel }
local channels = {}

function Channel:new()
    return setmetatable(
    {
    }, mt)
end

function Channel:Enable()
    self.channelEncoded = PrivateChat:GenerateID(self.db.name)
    channels[self.channelEncoded] = self
    --PrivateChat:Print("generated " .. self.channelEncoded )
    self:AddOptions()
    self:Register()
end

function Channel:SetupSlash()
    if _G["SLASH_"..self.channelEncoded.."1"] or not self.db.slash or self.db.slash =="" then
        return false
    else
        _G["SLASH_" .. self.channelEncoded .. "1"] = "/"..self.db.slash
        --PrivateChat:Print("just registered slash " .. "SLASH_" .. self.channelEncoded .. "1 = " ..  "/"..self.db.slash )
        return true
    end
end

function Channel:RemoveSlash()
        --PrivateChat:Print("about to kill slash " .. "SLASH_" .. self.channelEncoded .. "1 = " ..  "/"..self.db.slash )
    _G["SLASH_"..self.channelEncoded.."1"]=nil

    for i=1,NUM_CHAT_WINDOWS do
        if _G["ChatFrame"..i].editBox:GetAttribute("chatType") == self.channelEncoded then
            --PrivateChat:Print("Resetting sticky on frame " .. i )
            _G["ChatFrame"..i].editBox:SetAttribute("chatType","SAY")
            _G["ChatFrame"..i].editBox:SetAttribute("stickyType","SAY")
        end
    end
    hash_ChatTypeInfoList = {} -- very necessary. they made it 
end

function Channel:ChangeSlash(val)
    self:RemoveSlash()
    self.db.slash = val
    self:SetupSlash()
end

function Channel:ChangeFrame(checked, frame)
    -- yark, brute force lookup the global identifier of the frame
    for i=1,NUM_CHAT_WINDOWS do
        local globalFrameID = "ChatFrame"..i
        if _G[globalFrameID] == frame then
            self.db.frames[globalFrameID] = checked
        end
    end
end

function Channel:Register()
    local channelEncoded = self.channelEncoded
    if PrivateChat.registered[channelEncoded] then return end

    if self.db.name ~= "" and self.db.display ~= "" and self.db.password ~= "" then
        if not self:SetupSlash() then
            --PrivateChat:Print("setup slash fail?")
            
        end

        --PrivateChat:Print("Attempting to enable " .. self.db.name )
        _G["CHAT_MSG_" .. channelEncoded] = self.db.display
        _G["CHAT_" .. channelEncoded .. "_SEND"] = self.db.display .. ": "
        _G["CHAT_" .. channelEncoded .. "_GET"] = "%s"

        local info = { sticky = 1, flashTab = false, flashTabOnGeneral = false, r=self.db.r, g=self.db.g, b=self.db.b } 
        local infoCopy = {} 
        PrivateChat:tcopy(infoCopy,info)

        ChatTypeInfo[channelEncoded] = infoCopy
        ChatTypeGroup[channelEncoded] = { "CHAT_MSG_" .. channelEncoded }

        local configTable =
        {
            text="PrivateChat - " .. self.db.display,
            type=channelEncoded,
            noClassColor=1,
            checked = function () return IsListeningForMessageType(channelEncoded) end,
            func = 
            function (self, checked) 
                ToggleChatMessageGroup(checked, channelEncoded)
                local chan = PrivateChat:LookupChannelByEncodedID(channelEncoded)
                chan:ChangeFrame(checked, FCF_GetCurrentChatFrame())
            end
        } 
        table.insert(CHAT_CONFIG_CHAT_LEFT, 
            configTable
        )
        --CHAT_CATEGORY_LIST.PC = channelEncoded

        --table.insert(ChannelMenuChatTypeGroups, id)

        --for _, f in ipairs(self.db.profile.channels[id].frames) do
        --    table.insert(_G[f].messageTypeList, id)
        --end
        --

        -- forcibly insert ourselves into Frame 1
        --table.insert(_G["ChatFrame1"].messageTypeList, channelEncoded )

        if self.db.frames then
            for frameID, checked in pairs(self.db.frames) do
                if checked then
                    ChatFrame_AddMessageGroup(_G[frameID], channelEncoded )
                end
            end
        else
            self.db.frames = {}
        end

        self:Broadcast("CJ", "", channelEncoded) 
        
        PrivateChat:RiseChatEvent(channelEncoded, ": " .. L["joined channel."], UnitName("player"), self.db.display)
        PrivateChat.registered[channelEncoded] = true 

        ChatConfig_CreateCheckboxes(ChatConfigChatSettingsLeft, CHAT_CONFIG_CHAT_LEFT, "ChatConfigCheckBoxWithSwatchAndClassColorTemplate", PLAYER_MESSAGES);
    else
        --PrivateChat:Print("Not enabling " .. self.db.name .. " due to laziness")
    end
end

function Channel:Unregister()
    local channelEncoded = self.channelEncoded
    
    self:Broadcast("CL", "", self.channelEncoded) 

    PrivateChat:RiseChatEvent(self.channelEncoded, ": " .. L["left channel."], UnitName("player"), self.db.display)
    
    self:RemoveSlash()
    
    PrivateChat.registered[self.channelEncoded] = false

    for i=1,NUM_CHAT_WINDOWS do
        ChatFrame_RemoveMessageGroup(_G["ChatFrame"..i], channelEncoded )
    end

    _G["CHAT_" .. self.channelEncoded .. "_SEND"] = nil
    _G["CHAT_" .. self.channelEncoded .. "_GET"] = nil
    _G["SLASH_" .. self.channelEncoded .. "1"] = nil

    ChatTypeInfo[self.channelEncoded] = nil
    ChatTypeGroup[self.channelEncoded] = nil

    for key,value in pairs(CHAT_CONFIG_CHAT_LEFT) do
        if value.type==channelEncoded then
            table.remove(CHAT_CONFIG_CHAT_LEFT,key)
        end
    end
end

function Channel:Destroy()
    self:RemoveOptions()
    self:Unregister()
end

function Channel:RemoveOptions()
	PrivateChat.options.args.channels.args[self.db.name]=nil
end

function Channel:AddOptions()
    local db = self.db
    local options = 
    {
        type="group",
        name=function() return db.name end,
        args = 
        {
            rename = 
            {
                type = "input",
                name = "Channel name",
                desc = "Name of the channel",
                get=function() return db.name end,
                set=function(info,val) 
                    if val ~= db.name and val ~= "" then 
                        PrivateChat:RenameChannel(self,val)
                    end 
                end
            },
            display = 
            {
                type = 'input',
                name = 'Display name',
                desc = 'Name to show in your channel',
                get=function() return db.display end,
                set=function(info,val) if val ~= db.display and val ~= "" then db.display=val end end
            },
            password = 
            {
                type = 'input',
                name = 'Password',
                desc = 'Channel key/password',
                get=function() return db.password end,
                set=function(info,val) 
                    if val ~= db.password and val ~= "" then 
                        db.password=val 
                        local id = PrivateChat:GenerateID(db.name)
                        local chan = PrivateChat:LookupChannelByEncodedID(id)
                        chan:Register()
                    elseif val == "" then
                        db.password = ""
                        local id = PrivateChat:GenerateID(db.name)
                        local chan = PrivateChat:LookupChannelByEncodedID(id)
                        chan:Unregister()
                    end 
                end
            },
            slash = 
            {
                type = 'input',
                name = 'Slash Command',
                desc = 'Enter a slash command shortcut to access your channel. Omit the leading /',
                get=function() return db.slash end,
                set=function(info,val) 
                    if val ~= db.slash then 
                        local id = PrivateChat:GenerateID(db.name)
                        local chan = PrivateChat:LookupChannelByEncodedID(id)
                        chan:ChangeSlash(val)
                        db.slash=val
                    end 
                end
            },
            who =
            {
                type = 'execute',
                name = 'who',
                desc = 'Perform a check of who is in this channel',
                func = function(info) 
                    PrivateChat:ChannelWho(PrivateChat:GenerateID(db.name))
                end
            },
            descript = 
            {
                type = "description",
                name = function()
                    local id = PrivateChat:GenerateID(db.name)
                    if PrivateChat.registered[id] then
                        if db.slash == "" then
                            return "No slash command. Please add one or you will not be able to use your channel."
                        else
                            return "Your channel is activated"
                        end
                    else
                        return "This channel is currently inactive. Please ensure that all fields above are complete."
                    end
                end
            }
        }
    }
    PrivateChat.options.args.channels.args[self.db.name]=options
end


PrivateChat.defaults = 
{
    profile =
    {
        channels = {}
    }
}

PrivateChat.channelDefaults =
{
    name = 'NewChannel',
    display = 'MyNewChannel',
    password = ''
}

PrivateChat.options = {
    name = 'PrivateChat',
    type = 'group',
    args = {
        create = 
        {
            type="input",
            name = "Create channel",
            desc = "Create a new chat channel",
            set=
                function(info,val) 
                    if val and val ~= "" then 
                        PrivateChat:CreateChannel(val)
                    end 
                end,
            order=1
        },
        header = 
        {
            type="header",
            name="",
            order=3
        },
        channels = 
        {
            type="group",
            name="channels",
            desc="channels",
            args = {}
        },
        delete = 
        {
            type="select",
            name="Delete channel",
            desc="Choose the channel to be deleted",
            values=	function()
                local channels = {}
                for i, chan in pairs(PrivateChat:GetChannels()) do
                    channels[chan.db.name] = chan.db.name
                end
                return channels
            end,
            get=function() return deletechan end,
            set=function(self, val) deletechan = val end,
            order=4,
        },
        deleteexecute = 
        {
            type="execute",
            name="Delete channel",
            desc="Deletes the chosen channel",
            func=function(self) if deletechan then PrivateChat:DeleteChannel(deletechan) end end,
            order=5,
        },
        version = {
            type="execute",
            name="Version query",
            desc="Check others' versions",
            func=function(self) PrivateChat:VersionQuery() end,
            order=2,
        }

    }
}

-- +1 here realizing that that the tagged version will always be +1 vs the trunk version
-- <3
PrivateChat.revision = tostring(("$Revision: 60 $"):match("%d+") + 1)

PrivateChat.prefix = "PrivateChat" .. PROTOCOL_VERSION

function PrivateChat:tcopy(to, from)
  for k,v in pairs(from) do
    if(type(v)=="table") then
      to[k] = {}
      PrivateChat:tcopy(to[k], v);
    else
      to[k] = v;
    end
  end
end

function PrivateChat:PrintTable(table, depth)
    for i,v in pairs(table) do 
        self:Print(depth .. i .." - ") 
        if( type(v) == "string" ) then
            self:Print(depth .. " - " .. v) 
        elseif( type(v) == "number" ) then
            self:Print(depth .. " - " .. tostring(v))
        elseif( type(v) == "table" ) then
            PrivateChat:PrintTable(v,depth.."   ")
        else
            self:Print(depth .. "not sure how to print type: " .. type(v) )
        end

    end
end

function PrivateChat:GetChannels()
    return channels
end

function PrivateChat:CreateChannel(value, db)
    -- 
    --self:Print("Starting up channel " .. value)
    
	for i, chan in pairs(channels) do
		if chan.db.name == value then
            self:Print("Cannot create channel " .. value .. " due to conflict with existing channel.")
            return
        end
    end


    local channel = Channel:new()
    if not db then
        db = {}
        self:tcopy(db, PrivateChat.channelDefaults)

        channel.db = db
        channel.db.name = value
        channel.db.display = value
        channel.db.slash = value
        channel.db.password = ""
        channel.db.r = 1
        channel.db.g = 1
        channel.db.b = 1
        channel.db.frames = {}
        for i=1,NUM_CHAT_WINDOWS do
            channel.db.frames["ChatFrame"..i] = true
        end
    else
        channel.db = db
    end

    table.insert(self.db.profile.channels, db)
    channel:Enable()
end

function PrivateChat:DeleteChannel(name)
    local channelID = PrivateChat:LookupChannelByEncodedID(name)
	for i, chan in pairs(channels) do
		if chan.db.name == name then
            --self:Print("Destroying 1! " .. name .. "i= " .. i )
			chan:Destroy()
            channels[i] = nil
			--wipe(table.remove(channels, i))
		end
	end
	for i, chan in pairs(self.db.profile.channels) do
        if not chan.db then
            table.remove(self.db.profile.channels,i)
        elseif chan.db.name == name then
            --self:Print("Destroying 2! " .. name .. " i = " .. i )
			table.remove(self.db.profile.channels, i)
		end
	end
    deletechan = ""
end

function PrivateChat:RenameChannel(from, to)
    local dbTemp = from.db
    PrivateChat:DeleteChannel(from.db.name)
    from.db.name = to
    PrivateChat:CreateChannel(to,dbTemp)
end

-------------------------------------------------------------------------------
-- Main
-------------------------------------------------------------------------------

function PrivateChat:OnInitialize()
    --self.version = self.version .. " |cffff8888r".. self.revision .."|r" 
    self.db = LibStub("AceDB-3.0"):New("PrivateChatDB", self.defaults, "Default")

    if not self.db.profile.version or self.db.profile.version < DATABASE_VERSION then
        --self:Print("Resetting profile")
        self.db.profile.channels = {}
        self.db.profile.version = DATABASE_VERSION
    else
        --self:Print("Loading existing profile " .. self.db.profile.version .. "||||")
    end

    self.registered = {}

    self.options.args.profile = LibStub("AceDBOptions-3.0"):GetOptionsTable(self.db)
    LibStub("AceConfig-3.0"):RegisterOptionsTable("PrivateChat", self.options)
   	self.optionsFrame = LibStub("AceConfigDialog-3.0"):AddToBlizOptions("PrivateChat", "PrivateChat") 

    self:RegisterChatCommand("PrivateChat", "HandleCommand")
    self:RegisterChatCommand("pc", "HandleCommand")
    self:RegisterChatCommand("pchat", "HandleCommand")
end

function PrivateChat:HandleCommand(param)
    if param:sub(1,12) == "versionquery" then
        self:VersionQuery()
    elseif param:sub(1,7) == "version" then
        self:Print("Version: " .. self.revision)
    elseif param:sub(1,6) == "config" then
        InterfaceOptionsFrame_OpenToCategory("PrivateChat")
    elseif param:sub(1,3) == "who" then
        param = param:sub(4)
        local arg = self:GetArgs(param,1)
        if not arg or arg=="" then
            self:PrintUsage()
            return
        end
        local id = PrivateChat:GenerateID(arg)
        local channel = PrivateChat:LookupChannelByEncodedID(id)
        if not channel then
            self:Print("No such channel " .. arg )
        else
            self:ChannelWho(id)
        end
    else
        self:PrintUsage()
    end
end

function PrivateChat:PrintUsage() 
    self:Print("Usage:")
    self:Print("/pc version - print your version number")
    self:Print("/pc versionquery - print the versions of your guildmates")
    self:Print("/pc config - open the config panel")
    self:Print("/pc who *channel* - see who is in the channel")
end

function PrivateChat:OnEnable()
    self.routes = {}

    self:RawHook("SendChatMessage", true)
    self:RegisterComm(self.prefix) -- used to be in SetupTransportA
    self:SecureHook("ChangeChatColor")
    --self:SecureHook("ChatEdit_OnEditFocusLost")
    --self:RawHook("ChatEdit_UpdateHeader", true)

    -- create the channels object table from the (boring) data stored in db
    channels = {}
    for _, chan in pairs(self.db.profile.channels) do
        local channel = Channel:new()
        channel.db = chan
        channel:Enable()
    end
end
function PrivateChat:ChatEdit_UpdateHeader(eb)
	local type = eb:GetAttribute("chatType");
    --self:Print(type)
    self.hooks.ChatEdit_UpdateHeader(eb)
end
function PrivateChat:ChatEdit_OnEditFocusLost(fr)
--PrivateChat:PrintTable(fr, " ")
    local depth = ""
    for i,v in pairs(fr) do 
        self:Print(depth .. i .." - ") 
        if( type(v) == "string" ) then
            self:Print(depth .. " - " .. v) 
        elseif( type(v) == "number" ) then
            self:Print(depth .. " - " .. tostring(v))
        elseif( type(v) == "table" ) then
            --PrivateChat:PrintTable(v,depth.."   ")
        else
            self:Print(depth .. "not sure how to print type: " .. type(v) )
        end

    end

    self:Print("MOVING RIGHT ALONG")

    for i,v in pairs(fr.chatFrame.editBox) do 
        self:Print(depth .. i .." - ") 
        if( type(v) == "string" ) then
            self:Print(depth .. " - " .. v) 
        elseif( type(v) == "number" ) then
            self:Print(depth .. " - " .. tostring(v))
        elseif( type(v) == "table" ) then
            --PrivateChat:PrintTable(v,depth.."   ")
        else
            self:Print(depth .. "not sure how to print type: " .. type(v) )
        end

    end

end

function PrivateChat:ChangeChatColor(channelID,r,g,b) -- a hook
    if self.registered[channelID] then
        local channel = PrivateChat:LookupChannelByEncodedID(channelID)
        ChatTypeInfo[channelID].r = r
        ChatTypeInfo[channelID].g = g
        ChatTypeInfo[channelID].b = b

        -- save in DB too for restoration at load time. normally this would be in chat-cache.txt but
        -- blizz is a poo-head about fake channels
        channel.db.r = r
        channel.db.g = g
        channel.db.b = b

        PrivateChat:FireChatEvent("UPDATE_CHAT_COLOR", channelID, r, g, b)
    end
end

--function PrivateChat:OnDisable()
--    for id in pairs(self.db.profile.channels) do
--       self:UnregisterChannel(id)
--    end
--end

-------------------------------------------------------------------------------
-- Communication
-------------------------------------------------------------------------------

function PrivateChat:OnCommReceived(prefix, message, distribution, sender)
    if sender == UnitName("player") then return end -- filter off own messages (addon channel loops back your own stuff)
        
    local success, packet = self:Deserialize(message)

    if not success then return end
    --self:PrintTable(packet,"")
    local messageType = packet[1]
    if messageType == "VQ" then
        local challenge = packet[2]
        self:SendCommMessage(self.prefix, self:Serialize({"VR", challenge, self.revision}), "GUILD"); 
        return
    elseif messageType == "VR" then
        -- verify id, then display
        local challengeRemote = packet[2]
        if challengeRemote == self.challengeIssued then
            self:Print(sender .. " is using version " .. packet[3])
        end

        return
    end
    local channelID = packet[2]
    if not self.registered[channelID] then 
        return
    end
    local channel = PrivateChat:LookupChannelByEncodedID(channelID)
    if not channel then
        self:Print("Problem locating channel data for channel " .. channelID)
        return
    end
    local unencryptedPayload = packet[5]
    local hashRec = packet[3]
    local dataDec = AES:decrypt(channel.db.password, packet[4])
    local hashObs = self:StringHashDjb2(dataDec)
    if hashObs ~= hashRec then
        -- key mismatch
        --self:Print("Key mismatched message " .. channelID .. " " .. hashRec .. " " .. hashObs .. " from " .. sender)
        --return
    end

    local success, data

    if hashObs == hashRec then
        success, data = self:Deserialize(dataDec)
    else
        success, data = false, nil
    end
    --if not success then 
        --self:Print("Error deserializing :( ")
    --else
        --self:Print("Apparently I was successful deserializing")
    --end

    --if( type(data) == "table" ) then
    --    self:PrintTable(data,"")
    --else
    --    self:Print("I got " .. data )
    --end

    if messageType == "WQ" then
        -- encrypt challenge+ownName using key. return that. receiver can decrypt and verify the challenege
        -- is present and that it was sent by the sender. should be good verification, plus unique for each query and person
        if data[1] then
            local challengeRemote = data[1]
            channel:Broadcast("WR", unencryptedPayload, {challengeRemote, UnitName("player")})
        else
            channel:Broadcast("WR", unencryptedPayload, {0})
        end
    elseif messageType == "WR" then
        -- first, unencrypt the payload
        -- second, deserialize the interesting bit. check both the values in that packet

        -- verify the payload
        if data[1] and data[2] and data[1] == self.challengeIssued and data[2] == sender and unencryptedPayload==self.challengeIssued then
            -- good to go. display a message
            self:Print(sender.." is in channel " .. channel.db.name .. " with a matching key.")
        elseif unencryptedPayload == self.challengeIssued then
            self:Print(sender .. " is in channel " .. channel.db.name .. "with a mismatched key.")
        else
            -- do nothing. it must have been someone else who requested this.
        end
    elseif messageType == "CJ" then
        if data then
            local who = data

            --if channelInfo.alerts == 1 and not channelInfo.ignores[sender] then
            self:RiseChatEvent(channel.channelEncoded, ": " .. L["joined channel."], sender, channel.db.display)
            --end
        end
    elseif messageType == "CL" then
        --if channelInfo.alerts == 1 and not channelInfo.ignores[sender] then
            self:RiseChatEvent(channel.channelEncoded, ": " .. L["left channel."], sender, channel.db.display)
        --end
    elseif messageType == "CM" then
        --if not channelInfo.ignores[sender] then
        if data then
            self:RiseChatEvent(channel.channelEncoded, ": " .. data, sender, channel.db.display)
        end
    else
        self:Print("Received nonsense message " .. messageType)
    end
end

function Channel:Broadcast(command, unencryptedPayload, ...)
    if not PrivateChat:IsEnabled() then return end

    local rawPack = PrivateChat:Serialize(...)
    local hash = PrivateChat:StringHashDjb2(rawPack)
    local packet = AES:encrypt(self.db.password, rawPack)

    PrivateChat:SendCommMessage(PrivateChat.prefix, PrivateChat:Serialize({command, self.channelEncoded, hash, packet, unencryptedPayload}), "GUILD")
end

function PrivateChat:LookupChannelByEncodedID(channelEncoded)
    for i, chan in pairs(PrivateChat:GetChannels()) do
        --self:Print("is " .. channelEncoded .. " == " .. chan.channelEncoded )
        if channelEncoded==chan.channelEncoded then
            return chan
        end
    end
end

-------------------------------------------------------------------------------
-- Hooks
-------------------------------------------------------------------------------

-- raw hooked. if the incoming channel is a PC channel, then don't send via the normal method
function PrivateChat:SendChatMessage(text, chatType, language, channel)
    --self:Print( " | " .. chatType)
    if self.registered[chatType] then
        local channel = self:LookupChannelByEncodedID(chatType)

        --text = self:FilterTarget(channelPC, text)

        --self:RiseChatEvent(id, ": " .. self:FilterItemLinks(channelPC, text), UnitName("player"), channelPC.displayedName)
        self:RiseChatEvent(chatType, ": " .. text, UnitName("player"), channel.db.display)
        --channel:Broadcast("CM", chatType, 
        --    AES:encrypt(channel.db.password, text)
        --)
        channel:Broadcast("CM", "", text) 

        return
    else
        self.hooks["SendChatMessage"](text, chatType, language, channel)
    end
end

function PrivateChat:FireChatEvent(id, evt, a1,a2,a3,a4,a5,a6,a7,a8,a9,a10,a11)

    local bIsChat = (strsub(evt, 1,9)=="CHAT_MSG_");
    local chattype = strsub(evt, 10); -- might just be garbage if bIsChat==false, but in that case we don't use it

    for i=1,NUM_CHAT_WINDOWS do
        --    if(not bIsChat or MyAddOn_UserVars.Chat[chattype][i]) then
        local frame = _G["ChatFrame"..i]
        for _, v in pairs(frame.messageTypeList) do
            if (v == id) then
                ChatFrame_OnEvent(frame,evt,a1,a2,a3,a4,a5,a6,a7,a8,a9,a10,a11);
                --    end
            end
        end
    end
end

function PrivateChat:RiseChatEvent(id, message, sender, channel)
    local evt = "CHAT_MSG_" .. id;
    if sender == "" or message == "" or channel == "" then
        --self:Print( "Skipping Chat Event " .. message .. sender .. channel )
        return
    end
    self:FireChatEvent(tostring(id),evt,tostring(message),tostring(sender),"",tostring(channel),"","",0,"","","","");
end

-------------------------------------------------------------------------------
-- Channels
-------------------------------------------------------------------------------

function PrivateChat:GenerateID(name)
    local t = PrivateChat:MD5AsTable(string.lower(name))
    return strupper(string.format("%08x", bit.bxor(t[1], t[2], t[3], t[4])))
end

function PrivateChat:ValidateChannelName(name)
    if not string.find(name, "^[^%s+].*$") then
        return false
    end

    for _, channel in pairs(self.db.profile.channels) do
        if string.lower(channel.name) == string.lower(name) then
            return false
        end
    end

    return true
end

-------------------------------------------------------------------------------
-- Filtering
-------------------------------------------------------------------------------

--[[
function PrivateChat:FilterTarget(channel, text)
    return string.gsub(text, "%%t", UnitName("target") or L["<no target>"])
end

function PrivateChat:FilterItemLinks(channel, text)
    if channel.itemLinks == 1 then
        return text

    elseif channel.itemLinks == 2 then
        text = string.gsub(text, "(|c%x+|Hitem:%d+:%d+:%d+:%d+:%d+:%d+:%d+:%d+|h%[.-%]|h|r)", function(link)
            local _, _, itemString, name = string.find(link, "|c%x+|H(item:%d+:%d+:%d+:%d+:%d+:%d+:%d+:%d+)|h(%[.-%])|h|r")

            if GetItemInfo(itemString) then
                return link
            else
                return name
            end
        end)

    elseif channel.itemLinks == 3 then
        text = string.gsub(text, "(|c%x+|Hitem:%d+:%d+:%d+:%d+:%d+:%d+:%d+:%d+|h%[.-%]|h|r)", function(link)
            local _, _, name = string.find(link, "|c%x+|Hitem:%d+:%d+:%d+:%d+:%d+:%d+:%d+:%d+|h(%[.-%])|h|r")

            return name
        end)
    end

    return text
end
]]--

-------------------------------------------------------------------------------
-- Misc Functions
-------------------------------------------------------------------------------

function PrivateChat:StringHashDjb2(text)
    local hash, l, maxint = 5381, string.len(text), tonumber('ffffffff', 16)

    for i = 1, l do
        hash = bit.band(bit.lshift(hash, 5) + hash + string.byte(text, i), maxint)
    end

    return hash
end

function PrivateChat:VersionQuery()
    self:Print("Beginning version query")

    self.challengeIssued = math.random(1000000, 9999999)
    self:SendCommMessage(self.prefix, self:Serialize({"VQ", self.challengeIssued}), "GUILD"); 
end

function PrivateChat:ChannelWho(id)
    local chan = PrivateChat:LookupChannelByEncodedID(id)
    self:Print("Beginning who for channel " .. chan.db.name )

    self.challengeIssued = math.random(1000000, 9999999)
    chan:Broadcast("WQ", self.challengeIssued, {self.challengeIssued})
end

-------------------------------------------------------------------------------
-- MD5
-------------------------------------------------------------------------------
function PrivateChat:GenerateKey(key)
	return self:MD5AsTable(key)
end

local bit_bor, bit_band, bit_bxor = bit.bor, bit.band, bit.bxor
local bit_lshift, bit_rshift, bit_mod = bit.lshift, bit.rshift, bit.mod
local str_char, str_len, str_byte, str_sub = string.char, string.len, string.byte, string.sub
local str_rep, str_format = string.rep, string.format
local tinsert, tremove = table.insert, table.remove

local ff = tonumber('ffffffff', 16)
local consts = {
	3614090360, 3905402710, 606105819, 3250441966,
	4118548399, 1200080426, 2821735955, 4249261313,
	1770035416, 2336552879, 4294925233, 2304563134,
	1804603682, 4254626195, 2792965006, 1236535329,
	4129170786, 3225465664, 643717713, 3921069994,
	3593408605, 38016083, 3634488961, 3889429448,
	568446438, 3275163606, 4107603335, 1163531501,
	2850285829, 4243563512, 1735328473, 2368359562,
	4294588738, 2272392833, 1839030562, 4259657740,
	2763975236, 1272893353, 4139469664, 3200236656,
	681279174, 3936430074, 3572445317, 76029189,
	3654602809, 3873151461, 530742520, 3299628645,
	4096336452, 1126891415, 2878612391, 4237533241,
	1700485571, 2399980690, 4293915773, 2240044497,
	1873313359, 4264355552, 2734768916, 1309151649,
	4149444226, 3174756917, 718787259, 3951481745,
	1732584193, 4023233417, 2562383102, 271733878,
}

local f = function(x, y, z) return bit_bor(bit_band(x, y),bit_band(-x - 1, z)) end
local g = function(x, y, z) return bit_bor(bit_band(x, z),bit_band(y, -z - 1)) end
local h = function(x, y, z) return bit_bxor(x, bit_bxor(y, z)) end
local i = function(x, y, z) return bit_bxor(y, bit_bor(x, -z - 1)) end

local z = function(f, a, b, c, d, x, s, ac)
	a = bit_band(a + f(b, c, d) + x + ac, ff)
	return bit_bor(bit_lshift(bit_band(a, bit_rshift(ff, s)), s),bit_rshift(a, 32 - s)) + b
end

local function MD5Transform(X, A, B, C, D)

	local a, b, c, d = A, B, C, D

	a = z(f, a, b, c, d, X[ 0],  7, consts[ 1])
	d = z(f, d, a, b, c, X[ 1], 12, consts[ 2])
	c = z(f, c, d, a, b, X[ 2], 17, consts[ 3])
	b = z(f, b, c, d, a, X[ 3], 22, consts[ 4])
	a = z(f, a, b, c, d, X[ 4],  7, consts[ 5])
	d = z(f, d, a, b, c, X[ 5], 12, consts[ 6])
	c = z(f, c, d, a, b, X[ 6], 17, consts[ 7])
	b = z(f, b, c, d, a, X[ 7], 22, consts[ 8])
	a = z(f, a, b, c, d, X[ 8],  7, consts[ 9])
	d = z(f, d, a, b, c, X[ 9], 12, consts[10])
	c = z(f, c, d, a, b, X[10], 17, consts[11])
	b = z(f, b, c, d, a, X[11], 22, consts[12])
	a = z(f, a, b, c, d, X[12],  7, consts[13])
	d = z(f, d, a, b, c, X[13], 12, consts[14])
	c = z(f, c, d, a, b, X[14], 17, consts[15])
	b = z(f, b, c, d, a, X[15], 22, consts[16])

	a = z(g, a, b, c, d, X[ 1],  5, consts[17])
	d = z(g, d, a, b, c, X[ 6],  9, consts[18])
	c = z(g, c, d, a, b, X[11], 14, consts[19])
	b = z(g, b, c, d, a, X[ 0], 20, consts[20])
	a = z(g, a, b, c, d, X[ 5],  5, consts[21])
	d = z(g, d, a, b, c, X[10],  9, consts[22])
	c = z(g, c, d, a, b, X[15], 14, consts[23])
	b = z(g, b, c, d, a, X[ 4], 20, consts[24])
	a = z(g, a, b, c, d, X[ 9],  5, consts[25])
	d = z(g, d, a, b, c, X[14],  9, consts[26])
	c = z(g, c, d, a, b, X[ 3], 14, consts[27])
	b = z(g, b, c, d, a, X[ 8], 20, consts[28])
	a = z(g, a, b, c, d, X[13],  5, consts[29])
	d = z(g, d, a, b, c, X[ 2],  9, consts[30])
	c = z(g, c, d, a, b, X[ 7], 14, consts[31])
	b = z(g, b, c, d, a, X[12], 20, consts[32])

	a = z(h, a, b, c, d, X[ 5],  4, consts[33])
	d = z(h, d, a, b, c, X[ 8], 11, consts[34])
	c = z(h, c, d, a, b, X[11], 16, consts[35])
	b = z(h, b, c, d, a, X[14], 23, consts[36])
	a = z(h, a, b, c, d, X[ 1],  4, consts[37])
	d = z(h, d, a, b, c, X[ 4], 11, consts[38])
	c = z(h, c, d, a, b, X[ 7], 16, consts[39])
	b = z(h, b, c, d, a, X[10], 23, consts[40])
	a = z(h, a, b, c, d, X[13],  4, consts[41])
	d = z(h, d, a, b, c, X[ 0], 11, consts[42])
	c = z(h, c, d, a, b, X[ 3], 16, consts[43])
	b = z(h, b, c, d, a, X[ 6], 23, consts[44])
	a = z(h, a, b, c, d, X[ 9],  4, consts[45])
	d = z(h, d, a, b, c, X[12], 11, consts[46])
	c = z(h, c, d, a, b, X[15], 16, consts[47])
	b = z(h, b, c, d, a, X[ 2], 23, consts[48])

	a = z(i, a, b, c, d, X[ 0],  6, consts[49])
	d = z(i, d, a, b, c, X[ 7], 10, consts[50])
	c = z(i, c, d, a, b, X[14], 15, consts[51])
	b = z(i, b, c, d, a, X[ 5], 21, consts[52])
	a = z(i, a, b, c, d, X[12],  6, consts[53])
	d = z(i, d, a, b, c, X[ 3], 10, consts[54])
	c = z(i, c, d, a, b, X[10], 15, consts[55])
	b = z(i, b, c, d, a, X[ 1], 21, consts[56])
	a = z(i, a, b, c, d, X[ 8],  6, consts[57])
	d = z(i, d, a, b, c, X[15], 10, consts[58])
	c = z(i, c, d, a, b, X[ 6], 15, consts[59])
	b = z(i, b, c, d, a, X[13], 21, consts[60])
	a = z(i, a, b, c, d, X[ 4],  6, consts[61])
	d = z(i, d, a, b, c, X[11], 10, consts[62])
	c = z(i, c, d, a, b, X[ 2], 15, consts[63])
	b = z(i, b, c, d, a, X[ 9], 21, consts[64])

	return
		bit_band(A + a, ff),
		bit_band(B + b, ff),
		bit_band(C + c, ff),
		bit_band(D + d, ff)
end

local function LIntToString(i)
	return str_char(bit_band(bit_rshift(i, 0), 255)) .. str_char(bit_band(bit_rshift(i, 8), 255)) .. str_char(bit_band(bit_rshift(i, 16), 255)) .. str_char(bit_band(bit_rshift(i, 24), 255))
end

local function StringToBInt(s)
	local v = 0
	for i = 1, str_len(s) do
		v = v * 256 + str_byte(s, i)
	end
	return v
end

local function StringToLInt(s)
	local v = 0
	for i = str_len(s), 1, -1 do
		v = v * 256 + str_byte(s, i)
	end
	return v
end

local function StringToLIntArray(s, size, count)
	local o, r = 1, {}

	for i = 1, count do
		tinsert(r, StringToLInt(str_sub(s, o, o + size - 1)))
		o = o + size
	end

	return r
end

function PrivateChat:MD5AsTable(s)
	local l = str_len(s)
	local pad = 56 - bit_mod(l, 64)

	if bit_mod(l, 64) > 56 then
		pad = pad + 64
	end
	if pad == 0 then
		pad = 64
	end

	s = s .. str_char(128) .. str_rep(str_char(0), pad - 1)
	s = s .. LIntToString(8 * l) .. LIntToString(0)

	assert(bit_mod(str_len(s), 64) == 0)

	local a, b, c, d = consts[65], consts[66], consts[67], consts[68]

	for i = 1, str_len(s), 64 do
		local x = StringToLIntArray(str_sub(s, i, i + 63), 4, 16)

		assert(#(x) == 16)
		x[0] = tremove(x, 1)

		a, b, c, d = MD5Transform(x, a, b, c, d)
	end

	local swap = function (w)
		return StringToBInt(LIntToString(w))
	end

	return {
		swap(a),
		swap(b),
		swap(c),
		swap(d),
	}
end

